"""
Merqube API Session - subcomponent of the client library wrapper
"""

import json
import os
import uuid
from copy import deepcopy
from typing import Any, Optional, cast
from urllib.parse import urljoin

from cachetools import LRUCache, cached
from requests import PreparedRequest, Response, Session
from requests.adapters import HTTPAdapter
from requests.exceptions import HTTPError
from urllib3.util.retry import Retry

from merqube_client_lib.constants import (
    API_URL,
    MERQ_REQUEST_ID_ENV_VAR,
    REQUEST_ID_HEADER,
)
from merqube_client_lib.exceptions import APIError
from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.types import HTTP_METHODS
from merqube_client_lib.types import HTTPMethod as httpm

logger = get_module_logger(__name__)


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, timeout: int | None = None, **kwargs: Any):
        """
        Adapter to allow for setting timeouts on API calls.
        Timeout should be in seconds
        """
        super().__init__(**kwargs)
        self._timeout = timeout

    def send(self, request: PreparedRequest, **kwargs: Any) -> Response:  # type: ignore  # signature incompatible with supertype
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self._timeout

        return super().send(request, **kwargs)


class _RetrySession(Session):
    """abstraction over Session that provides retries"""

    def __init__(
        self,
        retries: int = 3,
        backoff_factor: float = 0.3,
        status_forcelist: tuple[int, ...] = (502, 504),  # status codes to retry on
        allowed_methods: list[str] = ["GET"],  # methods to retry on
        request_timeout: int | None = None,
    ):
        super().__init__()
        """
        By default, this session only retries on GET, but you may enable for DELETE, POST, and PUT as well. 
        Note that when considering idempotency, only the server state is considered, not the client state.
        See https://developer.mozilla.org/en-US/docs/Glossary/Idempotent

        In our APIs, DELETE, POST, and PUT are all idempotent with respect to the server state.

        However, from the client side, e.g., a lost ACK on a POST may result in a 409 DUPLICATE being returned.
        Similarly, a lost ACK on DELETE may result in a 404 NOT FOUND being returned.
        Finally, from the client side, for the majority of our APIs, two subsequent PUTs will not work due to the status key; you must do PUT GET PUT
        """
        for allowed in allowed_methods:
            assert allowed in HTTP_METHODS, f"Should be a valid http method: {', '.join(HTTP_METHODS)}"

        if allowed_methods:
            retry = Retry(
                total=retries,
                read=retries,
                connect=retries,
                backoff_factor=backoff_factor,
                status_forcelist=status_forcelist,
                allowed_methods=allowed_methods,
            )
            adapter = TimeoutHTTPAdapter(timeout=request_timeout, max_retries=retry)
            self.mount("http://", adapter)
            self.mount("https://", adapter)


class _BaseAPISession:
    """Base class for Merqube sessions"""

    def __init__(self, token: Optional[str] = None, **kwargs: Any):
        """
        Create a new merq session using token (the API Key)
        """
        self.session_args = kwargs
        self.token = token

        self._session: Optional[Session] = None
        self._session_pid: int = -1
        self._prefix_url = API_URL
        self.token_type = "APIKEY"

    @property
    def http_session(self) -> Session:
        """
        the http session, i.e. the requests.Session object

        The requests library isn't multiprocessing safe, but claims to be thread safe. So we can just ensure that
        the requests.Session object is only used in the process in which it was created
        """
        pid = os.getpid()
        if pid != self._session_pid or self._session is None:
            if self._session is not None:
                self._session.close()
            self._session = _RetrySession(**self.session_args)
            self._session_pid = pid

        return self._session

    def close(self) -> None:
        if self._session is not None:
            self._session.close()
            self._session = None
            self._session_pid = -1

    def __enter__(self):  # type: ignore
        """
        requests.Session like context manager behavior
        :return: the merq session
        """
        return self

    def __exit__(self, exc_type: str, exc_val: Any, exc_tb: Any) -> None:
        self.close()

    def request(
        self,
        method: httpm,
        url: str,
        params: dict[str, Any] | None = None,
        data: Any = None,
        headers: dict[str, str] | None = None,
        options: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        """
        Perform an http request with this session
        request session helpers sess.put, sess.get, etc, call this

        Note: urljoin has some perhaps non obvious behavior:

            In [2]: from urllib.parse import urljoin

            In [3]: urljoin("https://localhost:8080", "resource")
            Out[3]: 'https://localhost:8080/resource'

            In [4]: urljoin("https://localhost:8080", "/resource")
            Out[4]: 'https://localhost:8080/resource'

            In [5]: urljoin("https://localhost:8080/", "/resource")
            Out[5]: 'https://localhost:8080/resource'

            In [6]: urljoin("https://localhost:8080", "//resource")
            Out[6]: 'https://resource' # ??
        """
        assert params is None or options is None, "both params and options cannot be passed"
        if url.startswith("//"):
            logger.warning(f"URL {url} starts with //, this is probably a mistake")

        headers = deepcopy(headers) or {}  # do not modify client dict
        if self.token:
            headers["Authorization"] = f"{self.token_type} {self.token}"

        options_st = "" if not options else ("?" + "&".join([f"{k}={v}" for k, v in options.items()]))
        url = urljoin(self._prefix_url, url)
        logger.debug(f"Performing {method} on {url}{options_st}")
        options_dict: dict[str, str] = options or {}

        return self.http_session.request(
            method=method.value, url=url, params=options_dict or params, data=data, headers=headers, **kwargs
        )

    def get(self, url: str, **kwargs: Any) -> Response:
        """
        requests.Session like http GET method
        """
        return self.request(httpm.GET, url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> Response:
        """
        requests.Session like http PUT method
        """
        return self.request(httpm.PUT, url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> Response:
        """
        requests.Session like http PATCH method
        """
        return self.request(httpm.PATCH, url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> Response:
        """
        requests.Session like http POST method
        """
        return self.request(httpm.POST, url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> Response:
        """
        requests.Session like http DELETE method
        """
        return self.request(httpm.DELETE, url, **kwargs)


# Public


class MerqubeAPISession(_BaseAPISession):
    """
    API Session tailored to MerQube's API indexapi and its errors
    The session is initialized with an API Key, and that API Key is passed in the Authorization header for all requests
    """

    def __init__(
        self,
        token: str | None = None,
        **kwargs: Any,
    ):
        """can specify your own oauth token,default is no token (only public APIs)"""
        super().__init__(token=token, **kwargs)

    def handle_nonrecoverable(self, res: Response, exc: Exception, req_id: str) -> None:
        """helper function that handles a non recoverable error"""
        logger.exception(exc)
        try:
            # we may not have a json depending on the response
            rj = res.json()
        except (AttributeError, json.decoder.JSONDecodeError):
            rj = {}
        logger.debug(f"Request failed with status {res.status_code}: {rj}. Request ID: {req_id}")
        raise APIError(code=res.status_code, response_json=rj, request_id=req_id)

    def request_raise(self, method: httpm, url: str, **kwargs: Any) -> Response:
        """request method that logs the status code and raises on non 2XX"""

        # Generate a new requestid if this call isnt being made in a chain that already has it.
        # (eg client calls dataapi, dataapi calls secapi - we dont want the second call to overwrite the original)
        # if this isnt set by a client making a call to one of our APIs, the API itself will generate one (eg customer call)
        # order is: 1) explicitly specified 2) set via chain, 3) generate new
        headers = deepcopy(kwargs.pop("headers", {})) or {}
        headers[REQUEST_ID_HEADER] = (
            req_id := headers.get(REQUEST_ID_HEADER, os.getenv(MERQ_REQUEST_ID_ENV_VAR, str(uuid.uuid4())))
        )

        res = self.request(method=method, url=url, headers=headers, **kwargs)
        try:
            res.raise_for_status()
        except HTTPError as e:
            self.handle_nonrecoverable(res=res, exc=e, req_id=req_id)

        return res

    def get(self, url: str, **kwargs: Any) -> Response:
        return self.request_raise(httpm.GET, url, **kwargs)

    def put(self, url: str, **kwargs: Any) -> Response:
        return self.request_raise(httpm.PUT, url, **kwargs)

    def patch(self, url: str, **kwargs: Any) -> Response:
        return self.request_raise(httpm.PATCH, url, **kwargs)

    def post(self, url: str, **kwargs: Any) -> Response:
        return self.request_raise(httpm.POST, url, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> Response:
        return self.request_raise(httpm.DELETE, url, **kwargs)

    def get_json(self, url: str, options: Optional[dict[str, Any]] = None, **kwargs: Any) -> dict[str, Any]:
        """get where the result is json (as opposed to bytes for csv etc)"""
        return cast(dict[str, Any], self.get(url, options=options, **kwargs).json())

    def get_collection(self, url: str, options: Optional[dict[str, Any]] = None, **kwargs: Any) -> list[Any]:
        """get the inner results array from a collection API"""
        return cast(list[Any], self.get(url, options=options, **kwargs).json()["results"])

    def get_data(self, url: str, options: Optional[dict[str, Any]] = None, **kwargs: Any) -> str:
        """return raw data , most commonly used when options contains format=csv"""
        return self.get(url, options=options, **kwargs).content.decode().strip()

    def get_collection_single(self, url: str, options: Optional[dict[str, Any]] = None, **kwargs: Any) -> Any:
        """
        get the inner results array from a collection API when there is only one expected result
        for example, sess.get("index?name=XXX")
        """
        res = self.get_collection(url, options=options, **kwargs)
        num_results = len(res)
        if num_results != 1:
            raise APIError(400, {"message": f"Only one result was expected but {num_results} found!"})
        return res[0]


@cached(cache=LRUCache(maxsize=256))
def get_merqube_session(token: str | None = None, **session_args: Any) -> MerqubeAPISession:
    """
    Cached; returns a session with the given token
    """
    return MerqubeAPISession(token=token, **session_args)



"""
Merqube API constants
"""
from typing import Final

API_URL: Final[str] = "https://api.merqube.com"
REQUEST_ID_HEADER: Final[str] = "X-Request-ID"

MERQ_REQUEST_ID_ENV_VAR: Final[str] = "MERQ_REQUEST_ID"
MERQ_REQUEST_REMOTE_ADDR_ENV_VAR: Final[str] = "MERQ_REQUEST_REMOTE_ADDR"
MERQ_RUN_ID_ENV_VAR: Final[str] = "MERQ-RUN-ID"

DEFAULT_CACHE_TTL: Final[int] = 600



# generated by datamodel-codegen:
#   filename:  openapi.yaml
#   timestamp: 2023-05-31T20:26:09+00:00

from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field, StrictBool, StrictStr, conint, constr
from typing_extensions import Literal


class Role(Enum):
    data_point = "data point"
    verification = "verification"
    calculation = "calculation"
    development = "development"
    administration = "administration"


class Administrative(BaseModel):
    class Config:
        extra = Extra.forbid

    role: Role


class CalcFreq(Enum):
    Daily = "Daily"
    Daily__EOD = "Daily, EOD"
    Intraday = "Intraday"


class DisplayMonth(Enum):
    Jan = "Jan"
    Feb = "Feb"
    Mar = "Mar"
    Apr = "Apr"
    May = "May"
    Jun = "Jun"
    Jul = "Jul"
    Aug = "Aug"
    Sep = "Sep"
    Oct = "Oct"
    Nov = "Nov"
    Dec = "Dec"


class FieldModel(BaseModel):
    display_name: Optional[StrictStr] = None
    field_name: Optional[StrictStr] = None


class PortfolioDisplay(BaseModel):
    fields: Optional[List[FieldModel]] = None


class RebalFreq(Enum):
    Annual = "Annual"
    Daily = "Daily"
    Monthly = "Monthly"
    Quarterly = "Quarterly"
    Semi_Annual = "Semi-Annual"
    Weekly = "Weekly"
    Bi_Monthly = "Bi-Monthly"
    Intraday = "Intraday"
    None_ = "None"


class Stage(Enum):
    prod = "prod"
    test = "test"
    development = "development"


class WeightingMethod(Enum):
    Other = "Other"
    Equal = "Equal"


class MerqTimestamp(BaseModel):
    __root__: Union[
        datetime,
        date,
        constr(regex=r"^[0-9]{4}-[0-9]{2}-[0-9]{2}(?:[ T][0-9]{2}:[0-9]{2}:[0-9]{2})?(?:[.][0-9]{1,6})?$", strict=True),
    ] = Field(..., example=["2021-01-01", "2021-01-01T01:01:01", "2021-01-01T01:01:01.zzzzz"])


class Status(BaseModel):
    created_at: Optional[StrictStr] = None
    created_by: Optional[StrictStr] = None
    last_modified: StrictStr
    last_modified_by: Optional[StrictStr] = None
    locked_after: Optional[datetime] = Field(
        None,
        description="If this is set (non null), the manifest is locked for all edits to any other field after this timestamp. A PUT/PATCH may be used to first unlock the manifest, by setting this field (to a max of one hour in the future), or to `null` again, to make other edits.",
    )


class Condition(Enum):
    any = "any"
    all = "all"
    custom = "custom"


class WeekmaskEnum(Enum):
    Mon = "Mon"
    Tue = "Tue"
    Wed = "Wed"
    Thu = "Thu"
    Fri = "Fri"
    Sat = "Sat"
    Sun = "Sun"


class HolidayCalendarSpec(BaseModel):
    class Config:
        extra = Extra.forbid

    calendar_identifiers: Optional[List[constr(regex=r"^(FUT|MIC|FX):.+$", strict=True)]] = None
    swaps_monitor_codes: Optional[List[constr(regex=r"^[a-zA-Z]{3}$", strict=True)]] = None
    condition: Optional[Condition] = None
    holidays_to_add: Optional[List[date]] = None
    holidays_to_remove: Optional[List[date]] = None
    weekmask: Optional[List[WeekmaskEnum]] = Field(
        ["Mon", "Tue", "Wed", "Thu", "Fri"], description="weekmask of valid business days"
    )


class CommandArguments(BaseModel):
    __root__: List[List[StrictStr]]


class SubPod(BaseModel):
    class Config:
        extra = Extra.forbid

    arguments: Optional[CommandArguments] = None
    command: StrictStr
    pod_type: StrictStr
    uuid: StrictStr


class S3Bucket(BaseModel):
    class Config:
        extra = Extra.forbid

    bucket: StrictStr
    path: Optional[StrictStr] = None


class AirflowConfig(BaseModel):
    class Config:
        extra = Extra.forbid

    description: Optional[StrictStr] = Field(None, description="A short description related with the generated runner")
    s3_location: Optional[S3Bucket] = Field(
        None,
        description="Bucket and path where the generated dag should be stored. If not defined `merq-airflow-dags-airflow2` is defined in merqutil",
    )
    dag: Optional[Dict[str, Any]] = Field(None, description="use this to set kwargs for the dag")
    pod: Optional[Dict[str, Any]] = Field(None, description="use this to set kwargs for the pod")
    env: Optional[Dict[str, Any]] = Field(None, description="use this to set ENV variables in the pod")
    subpods: Optional[List[SubPod]] = None


class EmailDissapiConfig(BaseModel):
    recipient_list: List[StrictStr] = Field(..., min_items=1)
    email_list: Optional[List[StrictStr]] = None
    bcc_list: Optional[List[StrictStr]] = None
    cc_list: Optional[List[StrictStr]] = None
    subject: StrictStr
    content: Optional[StrictStr] = None
    file_keys: Optional[List[StrictStr]] = None


class SftpDissapiConfig(BaseModel):
    sftp_ids: List[StrictStr]
    file_keys: List[StrictStr]


class S3DissapiConfig(BaseModel):
    bucket: StrictStr
    key_prefix: StrictStr
    file_keys: List[StrictStr]


class DisseminationDestinations(BaseModel):
    email_dissapi_configs: Optional[List[EmailDissapiConfig]] = Field(None, min_items=1)
    sftp_dissapi_configs: Optional[List[SftpDissapiConfig]] = None
    s3_dissapi_configs: Optional[List[S3DissapiConfig]] = None


class IndexReport(BaseModel):
    uuid: StrictStr
    program_args: Optional[Dict[str, Any]] = None
    dissemination_destinations: Optional[DisseminationDestinations] = None


class RunConfigLabelAdditional(BaseModel):
    label: StrictStr = Field(..., description="unique label of this rc")


class HolidayCalendar(BaseModel):
    cal_type: Optional[StrictStr] = Field(None, example="MIC")
    mic: Optional[StrictStr] = Field(None, example="XNYS")


class Schedule(BaseModel):
    class Config:
        extra = Extra.forbid

    retries: int
    retry_interval_min: int
    schedule_start: StrictStr
    schedule_cron: Optional[StrictStr] = None
    timeout_delta: Optional[StrictStr] = Field(None, example="minutes=17")
    business_days_prior: Optional[int] = Field(None, example=1)
    holiday_calendar: Optional[HolidayCalendar] = None


class CopyFromManifest(BaseModel):
    copy_from_uuid: StrictStr = Field(..., description="the uuid of the resource to start from")
    field_deletes: Optional[List[StrictStr]] = Field(None, description="fields to delete from the from-manifest")
    field_updates: Optional[Dict[str, Any]] = Field(
        None,
        description="a dictionary of top level keys that the original manifest will be updated with (ie `manifest.update(this[field_updates])`. Top level keys can be added or replaced. To change the namespace, add namespace to the field_updates dict. The new manifest must be valid per the original resource definition.",
    )
    name: StrictStr = Field(..., description="the name of the new resource")


class CrudExtra(BaseModel):
    class Config:
        extra = Extra.forbid

    id: StrictStr
    namespace: StrictStr
    status: Status


class IndexClassDefinitionPost(BaseModel):
    class Config:
        extra = Extra.forbid

    name: StrictStr
    namespace: Optional[StrictStr] = None
    index_class_args_spec: Dict[str, Any]


class IndexClassDefinitionPatchPutGet(IndexClassDefinitionPost, CrudExtra):
    pass

    class Config:
        extra = Extra.forbid


class IndexSpec(BaseModel):
    index_class: Optional[StrictStr] = None
    index_class_args: Optional[Dict[str, Any]] = Field(
        None,
        description="should validate against index_class_args_spec field of the IndexClassDefinition identified by index_class",
    )
    index_variables: Optional[Dict[str, Any]] = Field(None, description="TODO")
    version: Optional[int] = None


class Type(Enum):
    security = "security"
    index = "index"


class Related(BaseModel):
    default_display: Optional[StrictBool] = False
    id: StrictStr
    metric: Optional[StrictStr] = None
    name: StrictStr
    type: Type


class ChartType(Enum):
    line = "line"
    spline = "spline"
    area = "area"
    bar = "bar"
    column = "column"


class CompareMode(Enum):
    absolute = "absolute"
    percent = "percent"


class ConstructorType(Enum):
    chart = "chart"
    stockChart = "stockChart"


class Value(BaseModel):
    class Config:
        extra = Extra.forbid

    display_name: StrictStr
    path: StrictStr


class Datum(BaseModel):
    class Config:
        extra = Extra.forbid

    url: StrictStr
    array_field: StrictStr
    values: List[Value]
    key_field: StrictStr


class CustomCharts(BaseModel):
    class Config:
        extra = Extra.forbid

    name: StrictStr
    xlabel: StrictStr
    ylabel: StrictStr
    chart_type: Optional[ChartType] = "line"
    compare_mode: Optional[CompareMode] = "percent"
    constructor_type: Optional[ConstructorType] = "stockChart"
    size: Optional[float] = 12
    data: Optional[List[Datum]] = None


class IndexBenchmark(BaseModel):
    class Config:
        extra = Extra.forbid

    ticker: StrictStr
    threshold: float
    index_metric: Optional[StrictStr] = Field(
        "price_return",
        description="the metric to compare against the ticker/threshold, eg can compare some other metric than price_return",
    )


class Provider(Enum):
    bloomberg = "bloomberg"
    reuters = "reuters"
    morningstar = "morningstar"
    factset = "factset"
    nasdaq = "nasdaq"
    wind = "wind"
    secapi = "secapi"


class BloombergProviderKwargsPostType(Enum):
    EOD = "EOD"
    RT = "RT"
    EOD_AND_RT = "EOD_AND_RT"


class BloombergProviderKwargs(BaseModel):
    class Config:
        extra = Extra.forbid

    post_type: BloombergProviderKwargsPostType


class ReutersProviderKwargsPostType(Enum):
    EOD = "EOD"
    RT = "RT"
    EOD_AND_RT = "EOD_AND_RT"


class ReutersProviderKwargs(BaseModel):
    class Config:
        extra = Extra.forbid

    post_type: Optional[ReutersProviderKwargsPostType] = None


class SecapiProviderKwargs(BaseModel):
    class Config:
        extra = Extra.forbid

    security_type: StrictStr
    security_name: StrictStr
    metric: StrictStr


class NasdaqEnv(Enum):
    testing = "testing"
    production = "production"


class NasdaqProviderKwargs(BaseModel):
    class Config:
        extra = Extra.forbid

    nasdaq_env: NasdaqEnv = Field(..., description="Which Nasdaq environment to send data to")
    eod_date_delta: Optional[int] = Field(
        None,
        description="How many days to offset the Nasdaq date (T if before 6:30pm EST otherwise T + 1) when getting EOD data from the Index",
    )


class Compression(Enum):
    NONE = "NONE"
    GZIP = "GZIP"


class LayoutType(Enum):
    SINGLE_FILE = "SINGLE_FILE"
    YEARLY_DIR_DAILY_FILE = "YEARLY_DIR_DAILY_FILE"


class DataCollectionsSpec(BaseModel):
    class Config:
        extra = Extra.allow

    index_dc_subtype: StrictStr = Field(
        ..., description="Index Data Collection Subtype object key moves to array after transformation"
    )
    auto_persist: Optional[StrictBool] = Field(
        False, description="Boolean to indicate whether to automatically persist the data. Default is False."
    )
    compression: Optional[Compression] = "GZIP"
    date_col: Optional[StrictStr] = "date"
    format: Optional[Literal["csv"]] = "csv"
    layout_type: Optional[LayoutType] = Field(
        "SINGLE_FILE",
        description="SINGLE_FILE: one file. YEARLY_DIR_DAILY_FILE: one directory per year, each day, a file with YYYYMMDD format.",
    )
    location: StrictStr
    name: constr(regex=r"^[A-Za-z0-9_]*$", min_length=3, max_length=128, strict=True)
    seed_file_path: Optional[StrictStr] = Field(None, description="Seed file for the data collection.")
    start_time: Optional[StrictStr] = None


class RealTimeTradeType(Enum):
    ELIGIBLE = "ELIGIBLE"
    INELIGIBLE = "INELIGIBLE"


class PortfolioUom(Enum):
    WEIGHT = "WEIGHT"
    UNITS = "UNITS"


class IdentifierType(Enum):
    RIC = "RIC"
    CURRENCY_CODE = "CURRENCY_CODE"


class AssetType(Enum):
    CASH = "CASH"
    THIRD_PARTY_INDEX = "THIRD_PARTY_INDEX"
    EQUITY = "EQUITY"


class SecurityCreationResponse(BaseModel):
    id: Optional[StrictStr] = None
    inserts: Optional[int] = None


class NewSecurity(BaseModel):
    class Config:
        extra = Extra.forbid

    name: StrictStr
    namespace: StrictStr = Field(
        ...,
        description="the calling user must have permission to this namespace to create the security in this namespace",
    )
    eff_ts: Optional[MerqTimestamp] = Field(
        None,
        description="DEPRECATED: this is no longer used for security creates, but allowed for backwards compatibility",
    )
    prov_ts: MerqTimestamp
    source: Optional[StrictStr] = None


class SecurityMetrics(BaseModel):
    eff_ts: MerqTimestamp
    id: StrictStr = Field(..., example="0000699a-fcd7-45d4-8ce4-064a4ffeced0")
    metric: StrictStr = Field(..., example="Level")
    source: Optional[StrictStr] = Field(None, example="some_gcp_feed")
    prov_ts: MerqTimestamp
    value: Union[StrictStr, float, int, StrictBool, List, Dict[str, Any]]


class SecurityMetricsDeletion(BaseModel):
    eff_ts: MerqTimestamp
    id: StrictStr = Field(..., example="0000699a-fcd7-45d4-8ce4-064a4ffeced0")
    metric: StrictStr = Field(..., example="Level")
    prov_ts: Optional[MerqTimestamp] = None
    source: Optional[StrictStr] = Field(None, example="some_gcp_feed")


class DataType(Enum):
    string = "string"
    float64 = "float64"
    datetime64 = "datetime64"
    int64 = "int64"
    bool = "bool"
    object = "object"


class MetricsSchema(BaseModel):
    data_type: DataType
    description: StrictStr = Field(..., description="a description of what this metric represents")
    name: StrictStr = Field(..., description="metric name")
    object_schema: Optional[Dict[str, Any]] = Field(
        None,
        description="optional - this supports complex objects, ie you can specify the  metric is of this schema. THis allows you to upload a json schema itself of this metric. THis is for downstream users of this metric.",
    )


class IdentifierType1(Enum):
    RIC = "RIC"
    secapi_name = "secapi_name"


class Security(BaseModel):
    class Config:
        extra = Extra.forbid

    identifier_type: IdentifierType1
    identifier_value: StrictStr = Field(..., example="AAPL.QQ")
    security_type: StrictStr = Field(..., example="equity")


class SecurityListPost(BaseModel):
    class Config:
        extra = Extra.forbid

    name: StrictStr = Field(..., description="unique name among all other security lists")
    namespace: Optional[StrictStr] = None
    realtime: Optional[StrictBool] = None
    securities: List[Security]


class SecurityListPatchPutGet(SecurityListPost, CrudExtra):
    pass

    class Config:
        extra = Extra.forbid


class OptionResponse(BaseModel):
    price: Optional[float] = Field(None, description="Option Price")
    vol: Optional[float] = Field(None, description="Implied Volatility")
    interest_rate: Optional[float] = Field(None, description="Implied Interest Rate")
    dividend_yield: Optional[float] = Field(None, description="Implied Dividend Yield for the underlying")
    spot: Optional[float] = Field(None, description="Underlying spot price")
    delta: Optional[float] = Field(None, description="Delta of this option")
    theta: Optional[float] = Field(None, description="Theta of the option being priced")
    gamma: Optional[float] = Field(None, description="Gamma of the option being priced")
    vega: Optional[float] = Field(None, description="Vega of the option being priced")
    bid_ask_spread: Optional[float] = Field(
        None, description="Spread between Bid and Ask prices of the option being priced"
    )
    request_id: Optional[StrictStr] = Field(None, description="Request ID sent in the request")


class FindStrikeResponse(BaseModel):
    strike: Optional[float] = Field(None, description="Strike that prices to the given budget.")
    spot: Optional[float] = Field(None, description="Underlying spot price.")
    strike_percentage: Optional[float] = Field(
        None, description="Strike as a percentage of the underlying spot price that prices to the given budget."
    )
    request_id: Optional[StrictStr] = Field(None, description="Request ID sent in the request")


class MetricsArrItem(BaseModel):
    id: Optional[StrictStr] = Field(None, example="price_return")
    value: Optional[float] = Field(None, example=1000.0000000000002)


class MetricsArr(BaseModel):
    __root__: List[MetricsArrItem]


class Stats(BaseModel):
    annual_volatility: Optional[float] = Field(None, example=3.717606464173326)
    annualized_return: Optional[float] = Field(None, example=6.665050717276233)
    cumulative_return: Optional[float] = Field(None, example=2.7597506251556547)
    id: Optional[StrictStr] = Field(None, example="YTD")
    max_drawdown: Optional[float] = Field(None, example=-1.3877029296510448)
    sharpe_ratio: Optional[float] = Field(None, example=1.759543308780235)
    start_date: Optional[MerqTimestamp] = None
    end_date: Optional[MerqTimestamp] = None


class MetricsDef(BaseModel):
    name: Optional[StrictStr] = Field(None, example="daily_return")
    type: Optional[StrictStr] = Field(None, example="double")


class ErrorCodes(BaseModel):
    code: Optional[StrictStr] = None
    message: Optional[StrictStr] = None


class IntradayTickFilter(BaseModel):
    class Config:
        extra = Extra.forbid

    start_date: Optional[MerqTimestamp] = None
    end_date: Optional[MerqTimestamp] = None
    start_time: Optional[StrictStr] = Field(
        None,
        description="Inclusive start time of the day to start running this intraday index. Format is %H:%M:%S.%f. This should be in the timezone of the index based on tzinfo.",
        example="20:39:00.262345",
    )
    end_time: Optional[StrictStr] = Field(
        None,
        description="Inclusive end time of the day to stop running this intraday index. Format is %H:%M:%S.%f. This should be in the timezone of the index based on tzinfo.",
        example="20:39:00.262345",
    )
    days_of_week: Optional[List[int]] = Field(
        None,
        description="Which days of the week to run the index. Each day of the week is denoted by an integer with Monday = 0 and Sunday = 6",
    )
    exclude_holidays: Optional[StrictBool] = Field(
        None,
        description="If this is True use the index holiday calendar to decide which days to run the index. It will only run on non-holidays.",
    )


class IntradayPublishConfigTargetBase(BaseModel):
    active_time_ranges: Optional[List[IntradayTickFilter]] = Field(
        None,
        description="Time ranges during which values should be published to this target. Target level active_time_ranges override publish_config metric level active_time_ranges.",
    )


class IntradayPublishConfigDbTarget(IntradayPublishConfigTargetBase):
    target: Literal["db"]


class IntradayPublishConfigSecapiTarget(IntradayPublishConfigTargetBase):
    target: Literal["secapi"]


class IntradayPublishConfigBloombergTargetParams(BaseModel):
    class Config:
        extra = Extra.forbid

    rounding_decimals: Optional[int] = None
    ticker: Optional[StrictStr] = None


class IntradayPublishConfigBloombergTarget(IntradayPublishConfigTargetBase):
    target: Literal["bloomberg"]
    params: Optional[IntradayPublishConfigBloombergTargetParams] = None


class IntradayPublishConfigReutersTargetParams(BaseModel):
    class Config:
        extra = Extra.forbid

    rounding_decimals: Optional[int] = None
    ticker: Optional[StrictStr] = None
    post_type: Optional[ReutersProviderKwargsPostType] = None


class IntradayPublishConfigReutersTarget(IntradayPublishConfigTargetBase):
    target: Literal["reuters"]
    params: Optional[IntradayPublishConfigReutersTargetParams] = None


class IntradayPublishConfigTarget(BaseModel):
    __root__: Union[
        IntradayPublishConfigDbTarget,
        IntradayPublishConfigSecapiTarget,
        IntradayPublishConfigBloombergTarget,
        IntradayPublishConfigReutersTarget,
    ]


class IntradayPublishConfigItem(BaseModel):
    class Config:
        extra = Extra.forbid

    default_active_time_ranges: Optional[List[IntradayTickFilter]] = Field(
        None,
        description="Time ranges during which values should be published for this metric. This will be overriden if a target level active_time_ranges is set.",
    )
    targets: List[IntradayPublishConfigTarget] = Field(..., description="List of targets to send this index metric too")


class IntradayPublishConfig(BaseModel):
    __root__: Optional[Dict[str, Union[IntradayPublishConfigItem, List[IntradayPublishConfigTarget]]]] = None


class Intraday(BaseModel):
    class Config:
        extra = Extra.forbid

    enabled: StrictBool = Field(..., description="If the intraday process should be run or not")
    security_id: Optional[StrictStr] = Field(
        None, description="Security ID of the intraday_index security where ticker data is served"
    )
    ticker_period: Optional[float] = Field(None, description="Duration between ticker datapoints. In seconds")
    tzinfo: Optional[StrictStr] = Field(None, description="Timezone of intraday tick timestamp comparisons")
    active_time_ranges: Optional[List[IntradayTickFilter]] = None
    publish_config: Optional[IntradayPublishConfig] = None
    calculation_max_retry_delay: Optional[conint(ge=0)] = Field(
        None,
        description="How many seconds to keep retrying the real time calculation before setting the force flag. Defalts to 0 which means retry forever.",
    )
    resource_cpu_request: Optional[conint(ge=1)] = Field(
        None, description="Minimum cpu resources requested in milliCPU"
    )
    resource_memory_request: Optional[conint(ge=0)] = Field(
        None, description="Minimum memory resources requested in mebibytes."
    )
    heartbeat_timeout: Optional[conint(ge=1)] = Field(
        None, description="Max number of seconds between heartbeats before the rtindex pod is considered unhealthy"
    )
    startup_failure_timeout: Optional[conint(ge=1)] = Field(
        None,
        description="Max number of seconds to wait for the first heartbeat from the rtindex pod before considering it a failed startup. By default it will wait 40sec",
    )
    data_refresh_period: Optional[conint(ge=0)] = Field(
        None,
        description="After the rtindex refreshes its data this is the number of seconds to wait before refreshing again. 0 means do not refresh. Defaults to 3600 seconds (60min)",
    )


class RunConfig(BaseModel):
    class Config:
        extra = Extra.allow

    airflow_config: Optional[AirflowConfig] = None
    command: Optional[StrictStr] = None
    command_arguments: Optional[CommandArguments] = None
    index_report_uuids: Optional[List[StrictStr]] = None
    index_reports: Optional[Union[List[StrictStr], List[IndexReport]]] = None
    job_enabled: StrictBool
    holiday_calendar: Optional[HolidayCalendarSpec] = None
    pod_image_and_tag: Optional[StrictStr] = Field(None, description="overrides the image")
    schedule: Schedule
    tzinfo: Optional[StrictStr] = None


class RunConfigLabel(RunConfig, RunConfigLabelAdditional):
    pass

    class Config:
        extra = Extra.allow


class Webpage(BaseModel):
    custom_charts: Optional[List[CustomCharts]] = None
    visible: Optional[StrictBool] = True


class IdentifierUUIDPost(BaseModel):
    class Config:
        extra = Extra.forbid

    display_name: Optional[StrictStr] = None
    index_name: StrictStr
    metric: Optional[StrictStr] = None
    name: StrictStr = Field(..., description="unique name among all other Identifiers of the provider resource type.")
    namespace: Optional[StrictStr] = None
    rounding_decimals: Optional[int] = None
    provider_kwargs: Optional[
        Union[BloombergProviderKwargs, ReutersProviderKwargs, SecapiProviderKwargs, NasdaqProviderKwargs]
    ] = None
    start_date: Optional[date] = None
    ticker: StrictStr


class IdentifierUUIDPatchPutGet(IdentifierUUIDPost, CrudExtra):
    pass

    class Config:
        extra = Extra.forbid


class IdentifierUUIDRef(BaseModel):
    class Config:
        extra = Extra.forbid

    name: StrictStr
    provider: Provider


class BasketPosition(BaseModel):
    class Config:
        extra = Extra.forbid

    amount: float = Field(..., title="Amount")
    asset_type: AssetType
    identifier: StrictStr = Field(..., title="Identifier")
    identifier_type: IdentifierType = Field(..., title="Identifier Type")
    position_id: Optional[StrictStr] = Field(None, title="Position Id")


class RicEquityPosition(BasketPosition):
    class Config:
        extra = Extra.forbid

    real_time_trade_types: Optional[List[RealTimeTradeType]] = None
    use_primary_listing: Optional[StrictBool] = Field(
        False,
        description="use the security's primary listing rather than the one from the identifier",
        title="Use Primary Listing",
    )


class Metrics(BaseModel):
    ts: Optional[datetime] = Field(None, description="iso8601 format")
    metrics: Optional[MetricsArr] = None


class IndexDefinitionPost(BaseModel):
    class Config:
        extra = Extra.forbid

    administrative: Administrative
    base_date: Optional[StrictStr] = None
    base_year: Optional[int] = None
    benchmark: Optional[Union[IndexBenchmark, List[IndexBenchmark]]] = None
    buffer_end: Optional[float] = None
    buffer_start: Optional[float] = None
    calc_freq: Optional[CalcFreq] = None
    currency: Optional[StrictStr] = None
    custom_metrics: Optional[List[MetricsSchema]] = None
    description: StrictStr
    display_month: Optional[DisplayMonth] = None
    documents: Optional[Dict[str, Any]] = None
    excess_strategy: Optional[Literal["keep_in_cash"]] = None
    family: StrictStr
    family_description: Optional[StrictStr] = None
    identifiers: Optional[List[IdentifierUUIDRef]] = Field(None, min_items=0)
    index_class_uuid: Optional[StrictStr] = None
    intraday: Optional[Intraday] = Field(
        None, description="Only available for indexes with intraday tick computation", title="Intraday Details"
    )
    launch_date: StrictStr
    methodology: Optional[StrictStr] = None
    month: Optional[int] = None
    name: constr(regex=r"^[a-zA-Z][a-zA-Z0-9-_]{0,99}$", strict=True)
    namespace: Optional[StrictStr] = None
    plot_metric: Optional[StrictStr] = None
    portfolio_allocation_display: Optional[StrictBool] = Field(
        None, description="whether or not portfolio_allocations are shown on the website. False by default"
    )
    portfolio_display: Optional[PortfolioDisplay] = None
    rebal_freq: Optional[RebalFreq] = None
    related: Optional[List[Related]] = None
    run_configuration: Optional[RunConfig] = None
    run_configurations: Optional[List[RunConfigLabel]] = None
    spec: Optional[IndexSpec] = None
    stage: Stage
    tags: Optional[StrictStr] = None
    title: StrictStr
    webpage: Optional[Webpage] = None
    weighting_method: Optional[WeightingMethod] = None


class IndexDefinitionPatchPutGet(IndexDefinitionPost, CrudExtra):
    pass

    class Config:
        extra = Extra.forbid


class EquityBasketPortfolio(BaseModel):
    class Config:
        extra = Extra.forbid

    positions: List[Union[BasketPosition, RicEquityPosition]] = Field(..., title="Positions")
    timestamp: MerqTimestamp = Field(
        ..., description="the time at which the portfolio should take effect", title="Timestamp"
    )
    unit_of_measure: PortfolioUom = Field(
        ...,
        description="indicates if the quantities in positions are absolute values, e.g. shares of equity positions, or relative weights of the total portfolio.",
    )



"""
Create an equity basket index
"""
from typing import Any, cast

import pandas as pd

from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.pydantic_types import IdentifierUUIDPost, IndexDefinitionPost
from merqube_client_lib.templates.equity_baskets.util import (
    create_index,
    load_template,
    read_file,
)

logger = get_module_logger(__name__)

TYPE_SPECIFIC_REQUIRED = ["constituents_csv_path", "base_value"]

# Optional keys that have defaults
TYPE_SPECIFIC_OPTIONAL = ["level_overrides_csv_path"]


def _get_constituents(base_date: str | pd.Timestamp, base_value: float, constituents_csv_path: str) -> dict[str, Any]:
    """read and validate constituents file"""
    constituents = read_file(constituents_csv_path)

    if pd.Timestamp(base_date) >= pd.Timestamp(min(c["date"] for c in constituents)):
        raise ValueError("base_date should be set to a date at least one day before the first portfolio date")

    inline = cast(
        dict[str, Any],
        [
            {
                "date": base_date,
                "identifier": "USD",
                "quantity": base_value,
                "security_type": "CASH",
            }
        ]
        + constituents,
    )

    return inline


def create(config_file_path: str, prod_run: bool = False) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    Creates a new Equity Basket with multiple entries
    This class does not handle Corax.
    """
    client, template, index_info, inner_spec = load_template(
        template_name="EQUITY_BASKET_TEMPLATE_V1",
        config_file_path=config_file_path,
        type_specific_req_fields=TYPE_SPECIFIC_REQUIRED,
        type_specific_opt_fields=TYPE_SPECIFIC_OPTIONAL,
    )

    ports = {
        "constituents": _get_constituents(
            constituents_csv_path=index_info["constituents_csv_path"],
            base_date=index_info["base_date"],
            base_value=index_info["base_value"],
        ),
        "date_type": "EFFECTIVE",
        "identifier_type": "RIC",
        "quantity_type": "SHARES",
        "specification_type": "INLINE",
    }

    inner_spec["portfolios"] = ports

    # TODO: do both and compare
    # TODO: on the target portfolio, set the specifcation type to API
    # target_portfolios = inline_to_tp(template)

    if pth := index_info.get("level_overrides_csv_path"):
        inner_spec["level_overrides"] = read_file(pth)

    return create_index(
        client=client, template=template, index_info=index_info, inner_spec=inner_spec, prod_run=prod_run
    )



"""
Merqube custom exceptions
"""

from typing import Any


class APIError(Exception):
    """Generic exception for when our client library hits a non 2xx code from our APIs"""

    def __init__(
        self, code: int | None = None, response_json: dict[str, Any] | None = None, request_id: str | None = None
    ):
        super().__init__()
        self.code = code
        self.response_json = response_json
        self.request_id = request_id



import json
import logging
from copy import deepcopy
from typing import Any, cast

import pandas as pd
import pytz
from pydantic.error_wrappers import ValidationError

from merqube_client_lib.api_client.merqube_client import MerqubeAPIClient
from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.pydantic_types import (
    BasketPosition,
    EquityBasketPortfolio,
    HolidayCalendarSpec,
    IdentifierUUIDPost,
    IdentifierUUIDRef,
    IndexDefinitionPatchPutGet,
    IndexDefinitionPost,
    IndexReport,
    IntradayPublishConfigBloombergTarget,
    IntradayPublishConfigTarget,
    PortfolioUom,
    Provider,
    RicEquityPosition,
    Stage,
)
from merqube_client_lib.pydantic_types import WeekmaskEnum as wm
from merqube_client_lib.util import freezable_utcnow

SPEC_KEYS = ["base_date"]
DATE_KEYS = ["base_date"]

# TODO: use jsonschema instead
TOP_LEVEL = ["namespace", "name", "title", "base_date", "description"]
MISC = ["apikey", "run_hour", "run_minute"]
REQUIRED_BASE_FIELDS = TOP_LEVEL + SPEC_KEYS + MISC

OPTIONAL_BASE_FIELDS = [
    "addl_comments",
    "bbg_ticker",
    "currency",
    "is_intraday",
    "timezone",
    "holiday_calendar",
    "email_list",
]


logger = get_module_logger(__name__, level=logging.DEBUG)


def inline_to_tp(index_model: IndexDefinitionPatchPutGet) -> list[tuple[str, EquityBasketPortfolio]]:
    """Convert the inline portfolio to a list of target portfolios"""

    target_portfolios = []

    assert index_model.spec and index_model.spec.index_class_args and index_model.spec.index_class_args.get("spec")
    portfolio = index_model.spec.index_class_args["spec"]["portfolios"]
    uom = PortfolioUom.UNITS if portfolio["quantity_type"] == "SHARES" else PortfolioUom.WEIGHT
    id_type = portfolio["identifier_type"]

    # see how nany TP values we need
    dates = set()
    for constituent in (const := portfolio["constituents"]):
        dates.add(constituent["date"])

    for date in sorted(dates):
        positions: list[BasketPosition | RicEquityPosition] = []
        for constituent in const:
            sec_type = constituent["security_type"]
            pos_class = RicEquityPosition if sec_type == "EQUITY" else BasketPosition

            if constituent["date"] != date:
                continue  # will get picked up in another target portfolio

            positions.append(
                pos_class(
                    asset_type=constituent["security_type"],
                    identifier=constituent["identifier"],
                    amount=constituent["quantity"],
                    # this is only at the top level of inline; but its wrong for the cash position (in TP it is CURRENCY_CODE)
                    identifier_type="CURRENCY_CODE" if sec_type == "CASH" else id_type,
                )
            )

        target_port_val = EquityBasketPortfolio(timestamp=date, unit_of_measure=uom, positions=positions)

        EquityBasketPortfolio.parse_obj(target_port_val)  # validate
        target_portfolios.append((date, target_port_val))

    return target_portfolios


def read_file(filename: str) -> list[Any]:
    """reads portfolio/overrides files"""
    return list(pd.read_csv(filename, float_precision="round_trip").to_dict(orient="index").values())


def log_index(index: IndexDefinitionPost) -> None:
    """
    Logs the resulting index
    """
    json_formatted_str = json.dumps(json.loads(index.json(exclude_none=True)), indent=4)
    logger.info("Index spec: \n" + json_formatted_str)


def get_index_info(
    config_file_path: str, type_specific_req_fields: list[str] = [], type_specific_opt_fields: list[str] = []
) -> dict[str, Any]:
    """
    load index specific info from config file + validate the data
    """
    index_info = cast(dict[str, Any], json.load(open(config_file_path, "r")))

    for k in (req := REQUIRED_BASE_FIELDS + type_specific_req_fields):
        if k not in index_info:
            raise ValueError(f"Missing required key {k} in index info")

    for k in index_info:
        if k not in req + OPTIONAL_BASE_FIELDS + type_specific_opt_fields:
            raise ValueError(f"Unknown key {k} in index info")

    try:
        for k in DATE_KEYS:
            pd.Timestamp(index_info[k])
    except ValueError as e:
        raise ValueError(
            "{k} must be a valid date. If the index should run at a certain time of day UTC, that should be part of the ISO8601 formatted first_run_date"
        ) from e

    if (tz := index_info.get("timezone")) and tz not in pytz.all_timezones:
        raise ValueError("Invalid timezone string: {tz}")

    if not (0 <= index_info["run_hour"] <= 23):
        raise ValueError("run_hour must be between 0 and 23")

    if not (0 <= index_info["run_minute"] <= 59):
        raise ValueError("run_minute must be between 0 and 59")

    # default: NYSE
    # TODO: we may have to allow configuration of the weekmask
    week = [wm.Mon, wm.Tue, wm.Wed, wm.Thu, wm.Fri]
    hol_cal = HolidayCalendarSpec(calendar_identifiers=["MIC:XNYS"], weekmask=week)
    if index_info.get("holiday_calendar"):
        try:
            assert isinstance(hc := index_info["holiday_calendar"], dict)
            if hc.get("mics"):
                hol_cal = HolidayCalendarSpec(calendar_identifiers=[f"MIC:{mic}" for mic in hc["mics"]], weekmask=week)
            elif hc.get("swaps_monitor_codes"):
                hol_cal = HolidayCalendarSpec(swaps_monitor_codes=hc["swaps_monitor_codes"], weekmask=week)
        except (AssertionError, ValidationError):
            raise ValueError("Invalid holiday calendar spec")

    index_info["_holiday_spec"] = hol_cal

    if index_info.get("email_list"):
        if not isinstance(index_info["email_list"], list) or not all(
            isinstance(x, str) for x in index_info["email_list"]
        ):
            raise ValueError("email_list must be a list of strings")

    return index_info


def load_template(
    template_name: str,
    config_file_path: str,
    type_specific_req_fields: list[str],
    type_specific_opt_fields: list[str] = [],
) -> tuple[MerqubeAPIClient, IndexDefinitionPost, dict[str, Any], dict[str, Any]]:
    """
    Loads a template index model to edit and then create a new index from
    """
    index_info = get_index_info(
        config_file_path=config_file_path,
        type_specific_req_fields=type_specific_req_fields,
        type_specific_opt_fields=type_specific_opt_fields,
    )
    token = index_info.get("apikey")

    client = MerqubeAPIClient(token=token)

    template = client.index_post_model_from_existing(template_name)

    assert template.spec and template.spec.index_class_args and template.spec.index_class_args.get("spec")
    inner_spec = template.spec.index_class_args["spec"]

    return client, template, index_info, inner_spec


def _configure_report(template: IndexDefinitionPost, email_list: list[str] | None = None) -> None:
    """
    confifugre index reports
    """
    updated_reports = []
    if not email_list or not template.run_configuration or not template.run_configuration.index_reports:
        logger.warning("Warning, index has no reports to configure or has not been given an email list")
        return

    for ind, report in enumerate(template.run_configuration.index_reports):
        updated_report = cast(IndexReport, deepcopy(report))
        assert updated_report.program_args
        diss_config = json.loads(updated_report.program_args["diss_config"].replace('\\"', '"'))

        if not (dc := diss_config.get("email")) or len(dc) > 1:
            logger.warning(f"Warning, report {ind} has no email config or has more than one config")
            continue

        diss_config["email"][0]["recipients"] = email_list
        updated_report.program_args["diss_config"] = json.dumps(diss_config).replace('"', '\\"')
        updated_reports.append(updated_report)
    template.run_configuration.index_reports = updated_reports


def configure_index(
    template: IndexDefinitionPost, index_info: dict[str, Any], inner_spec: dict[str, Any]
) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    set up the index from index info
    """
    for k in TOP_LEVEL:
        setattr(template, k, index_info[k] if k != "base_date" else pd.Timestamp(index_info[k]).strftime("%Y/%m/%d"))

    inner_spec["index_id"] = index_info["name"]
    for k in SPEC_KEYS:
        inner_spec[k] = index_info[k]
    inner_spec["holiday_spec"] = index_info["_holiday_spec"]

    name = index_info["name"]
    bbg_ticker = index_info.get("bbg_ticker")
    is_intraday = index_info.get("is_intraday")

    # set the start date at which this will start running
    assert template.run_configuration and template.run_configuration.schedule

    # set runtime
    # make sure to produce the last EOD value immediately on create
    today = freezable_utcnow() - pd.Timedelta(days=5)  # pyright: ignore
    template.run_configuration.schedule.schedule_start = pd.Timestamp(
        year=today.year,
        month=today.month,
        day=today.day,
        hour=index_info["run_hour"],
        minute=index_info["run_minute"],
        second=0,
    ).isoformat()
    template.run_configuration.tzinfo = index_info.get("timezone", "US/Eastern")

    template.stage = Stage.prod
    template.run_configuration.job_enabled = True

    if index_info.get("is_intraday"):
        assert template.intraday and template.intraday.publish_config
        template.intraday.enabled = True

    bbg_identifier_post = None
    if bbg_ticker:
        template.identifiers = [IdentifierUUIDRef(name=bbg_ticker, provider=Provider.bloomberg)]
        if is_intraday:
            assert template.intraday and template.intraday.publish_config and template.intraday.publish_config.__root__
            cast(list[IntradayPublishConfigTarget], template.intraday.publish_config.__root__["price_return"]).append(
                IntradayPublishConfigTarget(
                    __root__=IntradayPublishConfigBloombergTarget(target="bloomberg", active_time_ranges=None)
                )
            )

        bbg_identifier_post = IdentifierUUIDPost(
            name=bbg_ticker, index_name=name, namespace=index_info["namespace"], ticker=bbg_ticker
        )

    _configure_report(template=template, email_list=index_info.get("email_list"))

    return template, bbg_identifier_post


def create_index(
    client: MerqubeAPIClient,
    template: IndexDefinitionPost,
    index_info: dict[str, Any],
    inner_spec: dict[str, Any],
    prod_run: bool,
) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    Creates an index from a template
    """
    template, bbg_post = configure_index(template=template, index_info=index_info, inner_spec=inner_spec)

    log_index(template)

    if prod_run:
        # if we are posting to bloomberg, we need to create the identifier first
        if bbg_post:
            res = client.create_identifier(provider=Provider.bloomberg, identifier_post=bbg_post)
            logger.info(f"Created identifier: {res}")

        res = client.create_index(index_def=template)
        logger.info(f"Created index: {res}")

    return template, bbg_post



"""
Create an equity basket index
"""

import logging

from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.pydantic_types import IdentifierUUIDPost, IndexDefinitionPost
from merqube_client_lib.templates.equity_baskets.util import (
    create_index,
    load_template,
    read_file,
)

logger = get_module_logger(__name__, level=logging.DEBUG)

TYPE_SPECIFIC_REQUIRED = ["underlying_ric"]

# Optional keys that have defaults
TYPE_SPECIFIC_OPTIONAL = ["level_overrides_csv_path"]


def create(config_file_path: str, prod_run: bool = False) -> tuple[IndexDefinitionPost, IdentifierUUIDPost | None]:
    """
    Creates a new Equity Basket with multiple entries
    This class does not handle Corax.
    """
    client, template, index_info, inner_spec = load_template(
        template_name="SINGLE_STOCK_TR_TEMPLATE_VERSION_1",
        config_file_path=config_file_path,
        type_specific_req_fields=TYPE_SPECIFIC_REQUIRED,
    )

    inner_spec["portfolios"]["constituents"] = [
        {
            "date": index_info["base_date"],
            "identifier": index_info["underlying_ric"],
            "quantity": 1,
        }
    ]

    inner_spec["currency"] = index_info.get("currency", "USD")

    if pth := index_info.get("level_overrides_csv_path"):
        inner_spec["level_overrides"] = read_file(pth)

    return create_index(
        client=client, template=template, index_info=index_info, inner_spec=inner_spec, prod_run=prod_run
    )



"""
Create an equity basket index
"""
import logging
import os
from typing import Final

import click

from merqube_client_lib.logging import get_module_logger
from merqube_client_lib.templates.equity_baskets.decrement.create import (
    create as dec_index,
)
from merqube_client_lib.templates.equity_baskets.multiple_no_corax.create import (
    create as mult_index,
)
from merqube_client_lib.templates.equity_baskets.single_stock_total_return_corax.create import (
    create as ss_index,
)

SUPPORTED_INDEX_TYPES: Final[list[str]] = ["decrement", "single_stock_total_return", "multi_basket_no_corax"]

logger = get_module_logger(__name__, level=logging.DEBUG)


@click.command()
@click.option(
    "--index-type",
    type=click.Choice(SUPPORTED_INDEX_TYPES),
    required=True,
    help=f"type of index to create; must be one of {SUPPORTED_INDEX_TYPES}",
)
@click.option(
    "--config-file-path", type=str, required=True, help="path to the config file that follows the index template"
)
@click.option("--prod-run", is_flag=True, default=False, help="Create the index in production")
def main(index_type: str, config_file_path: str, prod_run: bool) -> None:
    """main entrypoint"""
    assert os.path.exists(config_file_path), f"Config file path does not exist: {config_file_path}"
    match index_type:
        case "decrement":
            func = dec_index
        case "single_stock_total_return":
            func = ss_index
        case "multi_basket_no_corax":
            func = mult_index
        case _:
            raise ValueError(f"Unknown index type: {index_type}")

    func(config_file_path=config_file_path, prod_run=prod_run)


if __name__ == "__main__":
    main()  # pyright: ignore  # pylint: disable=no-value-for-parameter
